/*------------------------------------------------------
 Maximum likelihood estimation 
 of migration rate  and effectice population size
 using a Metropolis-Hastings Monte Carlo algorithm                            
 -------------------------------------------------------                        
 R A N D O M   G E N E R A T O R   R O U T I N E S 
 
 creates options structures,
 reads options from parmfile if present
 
 prints options,
 and finally helps to destroy itself.
                                                                                                               
 Peter Beerli 1996, Seattle
 beerli@csit.fsu.edu
 
Copyright 1996-2002 Peter Beerli and Joseph Felsenstein, Seattle WA
Copyright 2003-2004 Peter Beerli, Tallahassee FL
 
 This software is distributed free of charge for non-commercial use
 and is copyrighted. Of course, we do not guarantee that the software
 works and are not responsible for any damage you may cause or have.
 
 
$Id: random.c,v 1.40 2004/08/28 19:59:34 beerli Exp $
-------------------------------------------------------*/
/* \file random.c */
#include "sighandler.h"
#include "migration.h"
#include "heating.h"
#include "random.h"
#include "migrate_mpi.h"

#ifdef SPRNG   /*NCSA random number generator library, slower than ours */
    #define SIMPLE_SPRNG
    #ifdef MPI
        #ifndef USE_MPI
            #define USE_MPI
        #endif
    #endif
    #include "sprng.h"
#endif

#ifdef MERSENNE_TWISTER
#include "mtwist-0.4/mtwist.h"
#endif

#ifdef DMALLOC_FUNC_CHECK
#include <dmalloc.h>
#endif

#ifdef PTHREADS
extern tpool_t heating_pool;
#endif

/* prototypes ----------------------------------------- */
void getseed (option_fmt * options);
void swap_ptr (long **ptr, long **newptr);

MYREAL randum (void);
#ifdef PTHREADS
MYREAL randum_thread (void);
#endif
long random_integer(long low, long high);


/// random integer
//=============================================
MYINLINE long random_integer(long low, long high)
{
    //Math.floor(Math.random()*(N-M+1))%(N-M+1)+M
    long r;
    long tt = high-low+1;
    r =  low + (long) (RANDUM() * tt);
    if(r<low || r>high)
        error("F.. [no bad words] but this IS shit\n");
    return r;
}

void
getseed (option_fmt * options)
{
#ifdef MPI
    long test=0;
#endif
#ifndef SPRNG
#ifndef MERSENNE_TWISTER
    long i;
#endif
#endif
    long timeseed;
    switch (options->autoseed)
    {
    case AUTO:
        timeseed = abs ((long) time (NULL));
#ifndef MERSENNE_TWISTER
        switch (timeseed % 4)
        {
        case 0:
            ++timeseed;
            break;
        case 1:
            break;
        case 2:
            ++timeseed;
            break;
        case 3:
            break;
        default:
            error ("Problem with automatic random number seed assignment");
        }
        options->inseed = abs ((long) timeseed);
#endif /*Mersenne-Twister*/
        break;
    case NOAUTO:
        break;
    case NOAUTOSELF:
        break;
    default:
        error ("Error: Seed value not defined");
        break;
    }
#ifndef SPRNG
#ifdef MPI
    MYMPIBCAST (&options->inseed, 1, MPI_LONG, MASTER, comm_world);
    if (myID != MASTER)
    {
#ifndef MERSENNE_TWISTER    
        for (i = 1; i <= 1000; i++) /* clear the random number generator */
            RANDUM ();

        for (i = 0; i <= 2; i++)
            seed[i] = 0;
        i = 0;
        do
        {
            seed[i] = options->inseed & 2047;
            options->inseed /= 2048;
            i++;
        }
        while (options->inseed != 0);
        for (i = 0; i <= myID; i++)
          {
            test = RANDINT (1, MAXLONG-1);
            //printf("     %i %li> %li\n", myID, i, test);
          }
        switch (test % 4)
          {
            case 0:
                test++;
                break;
            case 1:
                break;
            case 2:
                test--;
                break;
            case 3:
                test--;
                test--;
                break;
          }
        options->inseed = test;
    }
    FPRINTF (stdout, "%i> random number seed = %li\n", myID, options->inseed);
#else
        mt_seed32new((unsigned long)options->inseed);
#endif /*MERSENNE_TWISTER*/
#endif /*MPI*/
#endif /*SPRNG*/

    options->saveseed = options->inseed;
#ifdef SPRNG

    init_sprng (options->inseed, SPRNG_DEFAULT);
#else /* if we do not have SPRNG */

#ifndef MERSENNE_TWISTER
        for (i = 1; i <= 1000; i++) /* clear the random number generator */
            RANDUM ();
    for (i = 0; i <= 2; i++)
        seed[i] = 0;
    i = 0;
    do
    {
        seed[i] = options->inseed & 2047;
        options->inseed /= 2048;
        i++;
    }
    while (options->inseed != 0);
#else
    mt_seed32new((unsigned long) options->inseed);
#endif /*MERSENNE_TWISTER*/
#endif /* not SPRNG */
}

void
swap_ptr (long **ptr, long **newptr)
{
    long *temp;
    temp = *ptr;
    *ptr = *newptr;
    *newptr = temp;
}

#ifdef PTHREADS
MYREAL
randum_thread (void)
/* thread save random number generator */
{
#ifdef SPRNG
    return sprng ();
#else

    MYREAL value;
    //longer newseed;
#ifdef PTHREADS

    if ((pthread_mutex_lock (&(heating_pool->random_lock))) != 0)
        error ("pthread_mutex_lock failed in random_thread()");
#endif

    newseed[0] = 1549 * seed[0];
    newseed[1] = newseed[0] / 2048;
    newseed[0] &= 2047;
    newseed[2] = newseed[1] / 2048;
    newseed[1] &= 2047;
    newseed[1] += 1549 * seed[1] + 812 * seed[0];
    newseed[2] += newseed[1] / 2048;
    newseed[1] &= 2047;
    newseed[2] += 1549 * seed[2] + 812 * seed[1];
    swap_ptr (&newseed, &seed);
    //  memcpy (seed, newseed, sizeof (longer));
    seed[2] &= 1023;
    value = (((seed[0] / 2048.0 + seed[1]) / 2048.0 + seed[2]) / 1024.0);
#ifdef PTHREADS

    if ((pthread_mutex_unlock (&(heating_pool->random_lock))) != 0)
        error ("pthread_mutex_unlock failed in randum_thread()");
#endif

    return value;
#endif /*SPRNG*/
}    /* randum */
#else
MYREAL
randum (void)
/* Non thread-safe random number generator (faster) */
{
#ifdef SPRNG
    return sprng ();
#else
#ifdef MERSENNE_TWISTER
    return mt_drand();
#else
MYREAL value;
//  longer newseed;
newseed[0] = 1549 * seed[0];
newseed[1] = newseed[0] / 2048;
newseed[0] &= 2047;
newseed[2] = newseed[1] / 2048;
newseed[1] &= 2047;
newseed[1] += 1549 * seed[1] + 812 * seed[0];
newseed[2] += newseed[1] / 2048;
newseed[1] &= 2047;
newseed[2] += 1549 * seed[2] + 812 * seed[1];
//memcpy (seed, newseed, sizeof (longer));
swap_ptr (&newseed, &seed);
seed[2] &= 1023;
value = (((seed[0] / 2048.0 + seed[1]) / 2048.0 + seed[2]) / 1024.0);
return value;
#endif /*MERSENNE_TWISTER*/
#endif /*SPRNG*/
}    /* randum */
#endif
